/*
 * Title:        sag_internal/epl_plugins.hpp
 * Description:  Implementation of EPL plug-ins
 * $Copyright (c) 2015-2018, 2020 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: epl_plugin_wrap.hpp 366779 2020-02-05 10:42:04Z matj $
 */


#ifndef _SAG_INTERNAL_EPL_PLUGIN_WRAP_HPP_
#define _SAG_INTERNAL_EPL_PLUGIN_WRAP_HPP_

#include <sag_internal/cpp_stdlib.hpp>

/************************** Internal implementation of epl_plugin.hpp ************************************/
extern "C" {

/** 
 * @internal
 * Structure for passing method data across the C boundary
 */
struct ApamaEPLMethodData
{
	const char *name;
	char *signature;
	void *fn;
	bool blocking;
};

} // extern "C"

namespace com {
namespace apama {
namespace _EPLPLUGIN_INTERNAL_CPP_NAMESPACE { 

// Template classes for automatic type deduction of the method types.

template<typename T> struct _type_name;
template<> struct _type_name<double> { static std::string name() { return "float"; } };
template<> struct _type_name<bool> { static std::string name() { return "boolean"; } };
template<> struct _type_name<int64_t> { static std::string name() { return "integer"; } };
template<> struct _type_name<decimal_t> { static std::string name() { return "decimal"; } };
template<> struct _type_name<const char *> { static std::string name() { return "string"; } };
template<> struct _type_name<std::string> { static std::string name() { return "string"; } };
template<typename T> struct _type_name<custom_t<T> > { static std::string name() { return "chunk"; } };

template<typename T>
using type_name = _type_name<ampl::remove_const_t<ampl::remove_ref_t<T> > >;

template<typename T> struct get_return_value {static std::string type() { return " returns " + type_name<T>::name(); } };
template<> struct get_return_value<void> { static std::string type() { return ""; } };


template<typename... ARGS>
struct get_arg_types;
template<>
struct get_arg_types<> {
	static std::string args() { return ""; }
};
template<typename A>
struct get_arg_types<A> {
	static std::string args() 
	{
		// if you get an "incomplete type used in nested name specifier" here
		// it probably means you need to explicitly declare the type of your
		// EPL plug-in function because it can't be deduced from the arguments
		// or the type is not one of the supported types.
		// Deducible types are:
		// * int64_t
		// * double
		// * bool
		// * const char *
		// * std::string
		// * decimal_t
		// * const custom_t<T> &
		// Legal types which need explicit typing are:
		// * const list_t &
		// * const map_t &
		// * int64_t (when being passed a context)
		return type_name<A>::name(); 
	}
};
template<typename A, typename... ARGS>
struct get_arg_types<A, ARGS...>
{
	static std::string args() {
		return type_name<A>::name()+","+get_arg_types<ARGS...>::args();
	}
};

namespace detail {
/// casts a function pointer to a void* without warning.
/// WARNING: DRAGONS!
/// This is very non-portable. Fortunately, it's only as non-portable as the JNI layer,
/// which requires us to do this.
template<typename FN>
void* void_function_cast(FN *const &fnptr)
{
   return reinterpret_cast<void *const &>(fnptr);
} 
} // detail

/// classes to wrap a C++ member function as a C-linkage function and unpack/repack the arguments and return type
template<typename DERIVED, typename FN_T>
struct Wrap;

/** Specialization used to unpack the argument and return type */
template<typename DERIVED, typename RV, typename... ARGS>
struct Wrap<DERIVED, RV(DERIVED::*)(ARGS...)>
{
	typedef RV(DERIVED::*fn_t)(ARGS...); ///< type of the function being wrapped
	typedef data_t (of_t)(DERIVED *self, const list_t &args); ///< type of the wrapper
	
	/// Wrapper which takes a pointer to the plug-in argument and the list of arguments and returns a wrapped return value
	template <fn_t FN>
	static data_t of(DERIVED *self, const list_t &args);

	/// Wrapper which unwraps the argument types via an index tuple and variadic template expansion, but returns an unwrapped type
	template <fn_t FN, size_t... I>
	static RV call(DERIVED *self, const list_t &args, ampl::index_tuple<I...>)
	{
		return (self->*FN)(get<ARGS>(args[I])...);
	}
	/// This returns of<FN> cast to a void * without warning
	template<fn_t FN>
	static void *void_star_of()
	{
		return detail::void_function_cast<of_t>((of_t*) &of<FN>);
	}

	/** Return the automatically-deduced signature. If it's not deducible it'll fail to compile, 
	 * but if you don't call this (because you use the 2-arg version of registerMethod) then it
	 * won't try and compile it.
	 */
	static char *signature()
	{
		std::string sig = "action<"+get_arg_types<ARGS...>::args()+">" + get_return_value<RV>::type();
		return strdup(sig.c_str());
	}
};

/** This is used to wrap the return value back into a data_t. This is the general version
 * when the return type is non-void.
 */
template<typename RV, typename DERIVED, typename... ARGS>
struct GetReturn
{
	typedef RV(DERIVED::*fn_t)(ARGS...); ///< type of the function
	typedef Wrap<DERIVED, fn_t> wrap_t; ///< type of the wrap of the function
	/// Call the function using wrap_t's call, and then wrap the return value
	template<fn_t FN>
	data_t call(DERIVED *self, const list_t &args) const
	{
		return data_t(wrap_t::template call<FN>(self, args, typename ampl::make_index_tuple<sizeof...(ARGS)>::type()));
	}
};

/** And this is the version for a void return type */
template<typename DERIVED, typename... ARGS>
struct GetReturn<void, DERIVED, ARGS...>
{
	typedef void(DERIVED::*fn_t)(ARGS...); ///< type of the function
	typedef Wrap<DERIVED, fn_t> wrap_t; ///< type of the wrap of the function
	/// Call the function using wrap_t's call, and then return an empty result
	template<fn_t FN>
	data_t call(DERIVED *self, const list_t &args) const
	{
		wrap_t::template call<FN>(self, args, typename ampl::make_index_tuple<sizeof...(ARGS)>::type());
		return data_t();
	}
};

/** Take the wrapped args, and return wrapped return, using GetReturn and wrap_t::call */
template<typename DERIVED, typename RV, typename... ARGS>
template <RV(DERIVED::*FN)(ARGS...)>
inline data_t Wrap<DERIVED, RV(DERIVED::*)(ARGS...)>::of(DERIVED *self, const list_t &args)
{
	return GetReturn<RV, DERIVED, ARGS...>().template call<FN>(self, args);		
}


}}} // com.apama.epl

#endif // _SAG_INTERNAL_EPL_PLUGIN_WRAP_HPP_
